#!/usr/bin/ruby
# encoding: utf-8

require 'net/http'
require 'net/https'
require 'uri'
require 'hpricot' 
require 'timeout'

# Checked Links/URLs
$already_checked = Array.new
# Checked links that are not further stepped down
$checked_links = Array.new
# Failures, each Array element is a Hash consisting of blz, originating url, failed link and type of failure
$failed_links = Array.new
# Count the links checked for redirects per bank
$flat_links_blz = Hash.new
$flat_links_max = 50
# $cert_file = '/usr/share/curl/curl-ca-bundle.crt'
$cert_dir = '/etc/ssl/certs'

class SinglePage
	def initialize(url, depth, blz)
		#raise ArgumentError if url =~ /commerz/
		# $stderr.puts("DEPTH " + depth.to_f.to_s + " PARSING " + url )
		$stderr.puts("BLZ: " + blz.to_s + " PAGE: " + url + " (DEPTH " + depth.to_s + ") START PARSING")
		@url = url.strip
		@blz = blz.to_i
		@depth = depth.to_f
		$flat_links_blz[@blz] = 0 unless $flat_links_blz.has_key?(@blz)
		@redirect_to = nil
		# Framesets are treated as array of single pages
		@subpages = Array.new
		# Content is a plain string
		@pagecontent = nil
		# href/links that are followed are stored in an array (recursive data structure)
		@hrefs = Array.new
		# hrefs/links that point to external sites are stored in an array (flat data structure)
		@ext_sites = Array.new
		@guessed_url = false
		# store if https worked
		@https_ok = false
		unless @url =~ /^http/
			@url = "http://" + @url
			@guessed_url = true
		end
		# Force HTTPS first when starting
		if @depth < 0.05
			@url = @url.gsub('http://', 'https://')
			# $stderr.puts("Forcing HTTPS: " + @url)
			@guessed_url = true
		end
		# Push this URL to already checked URLs
		$already_checked.push(@url)
		@uri = URI.parse(@url)
		@host = @uri.host
		@proto = @uri.scheme
		@timeout = false
		@hdoc = nil
		@pagecontent = nil
		resp = nil
		begin
			if @proto == "http"
				Timeout::timeout(5) {
					resp = Net::HTTP.get_response(@uri)
					@pagecontent = resp.body
				}
			else
				# https_conn = Net::HTTP.new(@host, 443)
				# https_conn.use_ssl = true
				# https_conn.read_timeout = 3
				Timeout::timeout(3) {
					# $stderr.puts("Loading HTTPS: " + @url)
					# resp, body = Net::HTTP.get_response(@uri)
					https_conn = Net::HTTP.new(@uri.host, Net::HTTP.https_default_port)
					https_conn.use_ssl = true
					if File.directory?($cert_dir)
						https_conn.ca_path = $cert_dir
						https_conn.verify_mode = OpenSSL::SSL::VERIFY_PEER
						https_conn.verify_depth = 5
					end
					https_conn.ssl_timeout = 5
					# https.verify_mode = OpenSSL::SSL::VERIFY_PEER
					# https.ca_file = '/usr/share/curl/curl-ca-bundle.crt'
					# https.verify_depth = 2
					# https_conn.enable_post_connection_check = true
					https_conn.start do |http|
						path = @uri.path.to_s
						path = "/" if @uri.path.to_s == ""
						request = Net::HTTP::Get.new(path)
						resp = https_conn.request(request)
						@pagecontent = resp.body
					end
				}
			end
		rescue Timeout::Error
			resp = retry_after_fail
		rescue Errno::ECONNREFUSED
			resp = retry_after_fail
		rescue OpenSSL::SSL::SSLError
			resp = retry_after_fail
			# raise CertError
		rescue Errno::ECONNRESET
			resp = retry_after_fail
		rescue SocketError
			resp = retry_after_fail
		rescue Errno::EHOSTUNREACH
			resp = retry_after_fail
		end
		@https_ok = true if @url =~ /^https:/
		
		# $stderr.puts @pagecontent
		if resp.class.to_s == 'Net::HTTPRedirection' || 
			resp.class.to_s == 'Net::HTTPFound' || 
			resp.class.to_s == 'Net::HTTPMovedPermanently' || 
			resp.class.to_s == 'Net::HTTPTemporaryRedirect'
			$stderr.puts("BLZ: " + blz.to_s + " PAGE: " + url + " (DEPTH " + depth.to_s + ") -> " + resp['location'])
			# $stderr.puts(@url + " Forwards to: " + resp['location'])
			redir_loc = nil
			if resp['location'] =~ /^http/
				redir_loc = resp['location']
			elsif resp['location'] =~ /^\//
				redir_loc = @url + resp['location']
			else
				redir_loc = @url + "/" + resp['location']
			end
			@subpages.push(SinglePage.new(redir_loc, @depth + 0.3, @blz)) if @depth < 3.5
			# raise ArgumentError
		else
			if @timeout == false
				begin
				@hdoc = Hpricot.parse(@pagecontent)
				rescue
				# puts @pagecontent.to_s
				# puts resp.class
				# puts @uri.path
				# raise ArgumentError
				end
			end
		end
		# Check for framesets - nearly infinite depth!
		if !@pagecontent.nil? && depth < 1.0
			# @subpages =
			check_frameset
		end
		unless @pagecontent.nil? || @hdoc.nil?
			(@hdoc/:meta).each { |meta|
				if meta.attributes["http-equiv"] == "refresh"
					# puts meta.attributes["content"]
					tval = meta.attributes["content"].split(/;\s*URL=/)
					# puts tval[0]
					# puts tval[1]
					ltarg = check_hyperlinks( { "href" => tval[1].strip }, @url)
					aux_host = URI.parse(ltarg.strip).host if !ltarg.nil?
					aux_proto = URI.parse(ltarg.strip).scheme if !ltarg.nil?
					if !ltarg.nil? && depth < 1.5
						unless $already_checked.include?(ltarg)
							$stderr.puts("BLZ: " + blz.to_s + " PAGE: " + url + " (DEPTH " + depth.to_s + ") META-REDIR: -> " + ltarg)
							aux_page = SinglePage.new(ltarg.strip, depth + 0.4, @blz)
							@hrefs.push(aux_page)
						else
							# $stderr.puts("ALREADY CHECKED: " + ltarg)
							$stderr.puts("BLZ: " + blz.to_s + " PAGE: " + url + " (DEPTH " + depth.to_s + ") SKIP-META-REDIR: " + ltarg)
						end
					end
				end
			}
		end
		# Find images
		unless @pagecontent.nil? || @hdoc.nil?
			(@hdoc/:img).each { |img|
				$stderr.puts("BLZ: " + blz.to_s + " PAGE: " + url.to_s + " (DEPTH " + depth.to_s + ") IMAGE-FOUND: " + img.attributes["src"].to_s )
			}
		end
		# Find forms
		unless @pagecontent.nil? || @hdoc.nil?
			(@hdoc/:form).each { |form|
				$stderr.puts("BLZ: " + blz.to_s + " PAGE: " + url + " (DEPTH " + depth.to_s + ") FORM-FOUND: " + form.attributes["action"].to_s )
			}
		end
		
		# Check for all contained links - max depth one
		unless @pagecontent.nil? || @hdoc.nil?
			(@hdoc/:a).each { |link|
				begin
					ltarg = check_hyperlinks(link.attributes, @url)
					aux_host = URI.parse(ltarg.strip).host if !ltarg.nil?
					aux_proto = URI.parse(ltarg.strip).scheme if !ltarg.nil?
					if !ltarg.nil? && depth < 1.0 && aux_host == @host && aux_proto == @proto
						unless $already_checked.include?(ltarg)
							aux_page = SinglePage.new(ltarg.strip, depth + 1.0, @blz)
							@hrefs.push(aux_page)
						else
							# $stderr.puts("ALREADY CHECKED: " + ltarg)
							$stderr.puts("BLZ: " + blz.to_s + " PAGE: " + url + " (DEPTH " + depth.to_s + ") SKIP: " + ltarg)
						end
					elsif !ltarg.nil? && (aux_host != @host || aux_proto != @proto)
						# $stderr.puts("CHECKING EXTERNAL HOST!")
						# $stderr.puts("BLZ: " + blz.to_s + " PAGE: " + url + " (DEPTH " + depth.to_s + ") CHECKING_EXT... " + ltarg ) 
						aux_hash = Hash.new
						aux_hash["proto"] = URI.parse(ltarg.strip).scheme
						aux_hash["host"] = aux_host
						@ext_sites.push(aux_hash)
						# Check if a link redirects, check just depth one
						redir = nil
						begin
							redir = check_redirect(ltarg, 0)
							#raise WeirdError
						rescue
						end
						unless redir.nil?
							$checked_links.push(ltarg)
							aux_hash = Hash.new
							aux_hash["proto"] = URI.parse(redir.strip).scheme
							aux_hash["host"] = URI.parse(redir.strip).host
							# ignore links that target the same host
							if aux_hash["host"] != @host || aux_hash["proto"] != @proto
								# $stderr.puts("EXTERNAL LINK MASKING REDIRECT FOUND!")
								$stderr.puts("BLZ: " + blz.to_s + " PAGE: " + url + " (DEPTH " + depth.to_s + ") REDIR: " +
									ltarg + " -> " + redir ) 
								@ext_sites.push(aux_hash)
							end
						else 
							# $stderr.puts("LINK NOT MASKING REDIRECT, SKIP CONT!")
							$stderr.puts("BLZ: " + blz.to_s + " PAGE: " + url + " (DEPTH " + depth.to_s + ") SKIP/NOREDIR: " +
									ltarg ) 
						end
					elsif !ltarg.nil? && !$already_checked.include?(ltarg) && 
						!$checked_links.include?(ltarg) && $flat_links_blz[@blz] < $flat_links_max
						# Check if a link redirects, check just depth one
						redir = nil
						begin
							redir = check_redirect(ltarg, 0)
						rescue
						end
						unless redir.nil?
							$checked_links.push(ltarg)
							aux_hash = Hash.new
							aux_hash["proto"] = URI.parse(redir.strip).scheme
							aux_hash["host"] = URI.parse(redir.strip).host
							# ignore links that target the same host
							if aux_hash["host"] != @host || aux_hash["proto"] != @proto
								# $stderr.puts("EXTERNAL LINK MASKING REDIRECT FOUND!")
								$stderr.puts("BLZ: " + blz.to_s + " PAGE: " + url + " (DEPTH " + depth.to_s + ") REDIR: " +
									ltarg + " -> " + redir ) 
								@ext_sites.push(aux_hash)
								#puts "HULLABULLA"
							end
							#raise FuckingError
						else 
							# $stderr.puts("LINK NOT MASKING REDIRECT, SKIP CONT!")
							$stderr.puts("BLZ: " + blz.to_s + " PAGE: " + url + " (DEPTH " + depth.to_s + ") SKIP/NOREDIR: " + ltarg.to_s)
						end
						$flat_links_blz[@blz] += 1	
					else	
						$stderr.puts("BLZ: " + blz.to_s + " PAGE: " + url + " (DEPTH " + depth.to_s + ") MAXCOUNT/NIL: " + ltarg.to_s)
					end
				rescue URI::InvalidURIError
					# $stderr.puts("INVALID URL!")
					$stderr.puts("BLZ: " + blz.to_s + " PAGE: " + url + " (DEPTH " + depth.to_s + ") INVALID: " + ltarg.to_s)
				end
				
			}
		end
		# Exit if Frames found (debug only)
		if @subpages.size > 0
			# $stderr.puts("Frames found: " + @subpages.size.to_s)
			
			# raise ArgumentError
		end
	end
	attr_reader :url, :uri, :redirect_to, :subpages, :hrefs, :ext_sites, :already_checked, :https_ok

	# Extract all links 
	
	def check_hyperlinks(attrs, baseurl)
		begin
		if !attrs["onclick"].nil? || !attrs["onClick"].nil?
			onclick = attrs["onclick"] unless attrs["onclick"].nil? 
			onclick = attrs["onClick"] unless attrs["onClick"].nil? 
			$stderr.puts("BLZ: " + @blz.to_s + " PAGE: " + @url + " (DEPTH " + @depth.to_s + ") ONCLICK: " + onclick)
		end
		if attrs["href"] =~/^http:\/\// || attrs["href"] =~/^https:\/\//
			# $stderr.puts("LINK FULLURL/EXTERNAL: " + attrs["href"])
			# $stderr.puts("BLZ: " + @blz.to_s + " PAGE: " + @url + " (DEPTH " + @depth.to_s + ") FULLURL: " + attrs["href"])
			return attrs["href"]
		elsif attrs["href"] =~ /^\//
			# $stderr.puts("LINK ABSOLUTE/LOCAL: " + attrs["href"] + " DESCENDING")
			# $stderr.puts("BLZ: " + @blz.to_s + " PAGE: " + @url + " (DEPTH " + @depth.to_s + ") FULLURL: " + attrs["href"])
			return build_url_relative(URI.parse(@url), attrs["href"])
		elsif attrs["href"] =~ /^mailto:/
			# $stderr.puts("IGNORE EMAIL ADRESS ") #  + attrs["href"])
			$stderr.puts("BLZ: " + @blz.to_s + " PAGE: " + @url + " (DEPTH " + @depth.to_s + ") LINK_MAIL: " + attrs["href"])
			return nil
		elsif attrs["href"] =~ /^javascript:/i
			# $stderr.puts("IGNORE JAVASCRIPT LINK ") #  + attrs["href"])
			$stderr.puts("BLZ: " + @blz.to_s + " PAGE: " + @url + " (DEPTH " + @depth.to_s + ") LINK_JS: " + attrs["href"])
			return nil
		elsif attrs["href"] =~/^#/
			# $stderr.puts("LINK FAKE CHECK FOR JS")
			$stderr.puts("BLZ: " + @blz.to_s + " PAGE: " + @url + " (DEPTH " + @depth.to_s + ") LINK_FAKE: " + attrs["href"])
			return nil
		elsif !attrs["href"].nil?
			# $stderr.puts("LINK RELATIVE: " + attrs["href"] + " DESCENDING")
			return build_url_relative(URI.parse(@url), attrs["href"])
		end
		rescue URI::InvalidURIError
		end
		return nil
	end
	
	def retrieve_hostprotos
		resset = @ext_sites
		# Add this host/proto
		this_host = Hash.new
		this_host["proto"] = @uri.scheme
		this_host["host"] = @uri.host
		resset.push(this_host)
		# Add this redir host/proto
		# Add all subpages' ressets
		@hrefs.each { |h| resset = resset + h.retrieve_hostprotos }
		# Add all frames' ressets
		@subpages.each { |h| resset = resset + h.retrieve_hostprotos }
		# flatten
		return resset.uniq
	end
	
	def check_frameset
		toks = @pagecontent.split
		frame_found = false
		framesets = Array.new
		aux_frame = Array.new
		inside_frame = false
		toks.each { |t|
			# $stderr.puts(t)
			frame_found = true if t.upcase =~ /<FRAMESET/
			inside_frame = true if t.upcase =~ /<FRAME$/ 
			if inside_frame == true
				aux_frame.push(t.chomp)
				# $stderr.puts(t)
			end
			if t =~ />/ && inside_frame == true
				framesets.push(aux_frame)
				aux_frame = Array.new
				unless t.upcase =~ /<FRAME$/
					inside_frame = false
				else
					aux_frame.push(t)
				end
			end
		}
		if frame_found == true
			# $stderr.puts("FRAMESET FOUND IN: " + @uri.scheme + " + " + @uri.port.to_s + " + " + @uri.host + " + " + @uri.path )
			$stderr.puts("BLZ: " + @blz.to_s + " PAGE: " + @url + " (DEPTH " + @depth.to_s + ") FRAMES FOUND!")
			src_regexp = /[sS][rR][cC]=\"(.*?)\"/
			framesets.each { |f|
				# $stderr.puts f.join(" -+- ") # unless f.size < 1
				f.each { |t|
					match = src_regexp.match(t)
					unless match.nil?
						if match[1] =~ /^http:\/\// || match[1] =~ /^https:\/\//
							url_to_check = match[1]
							unless $already_checked.include?(url_to_check)
								# $stderr.puts "FRAME " + @depth.to_s + " FULLURL  " + match[1]
								$stderr.puts("BLZ: " + @blz.to_s + " PAGE: " + @url + " (DEPTH " + @depth.to_s + ") FRAME: " + match[1] )
								# raise ArgumentError
								@subpages.push(SinglePage.new(match[1], @depth + 0.4, @blz))
								$already_checked.push(match[1])
							else
								# $stderr.puts("ALREADY CHECKED: " + url_to_check)
								$stderr.puts("BLZ: " + @blz.to_s + " PAGE: " + @url + " (DEPTH " + @depth.to_s + ") KNOWNFRAME: " + match[1] )
							end
						elsif match[1] =~ /^\//
							url_to_check = build_url_relative(@uri, match[1])
							unless $already_checked.include?(url_to_check)
								# $stderr.puts "FRAME " + @depth.to_s + " ABSOLUTE " + match[1] + 
								#	" RESULTS IN " + build_url_relative(@uri, match[1])
								$stderr.puts("BLZ: " + @blz.to_s + " PAGE: " + @url + " (DEPTH " + @depth.to_s + ") FRAME: " + url_to_check )
								@subpages.push(SinglePage.new(build_url_relative(@uri, match[1]), @depth + 0.4, @blz))
								$already_checked.push(build_url_relative(@uri, match[1]))
							else
								# $stderr.puts("ALREADY CHECKED: " + url_to_check)
								$stderr.puts("BLZ: " + @blz.to_s + " PAGE: " + @url + " (DEPTH " + @depth.to_s + ") KNOWNFRAME: " + url_to_check )
							end
						else
							url_to_check = build_url_relative(@uri, match[1])
							unless $already_checked.include?(url_to_check)
								# $stderr.puts "FRAME " + @depth.to_s + " RELATIVE " + match[1] + 
								#	" RESULTS IN " + build_url_relative(@uri, match[1])
								$stderr.puts("BLZ: " + @blz.to_s + " PAGE: " + @url + " (DEPTH " + @depth.to_s + ") FRAME: " + url_to_check )
								@subpages.push(SinglePage.new(build_url_relative(@uri, match[1]), @depth + 0.4, @blz))
								$already_checked.push(build_url_relative(@uri, match[1]))
							else
								# $stderr.puts("ALREADY CHECKED: " + url_to_check)
								$stderr.puts("BLZ: " + @blz.to_s + " PAGE: " + @url + " (DEPTH " + @depth.to_s + ") KNOWNFRAME: " + url_to_check )
							end
						end
					end
				}
			}
			# $stderr.puts @pagecontent
		end
		return framesets
	end
	
	def build_url_relative(uri, link)
		if uri.path =~ /\/$/ && uri.path =~ /\?/
			aux_path = uri.path.split("?")
			uri.path = aux_path[0]
		end
		if link =~ /^\//
			return uri.scheme + "://" + uri.host + ":" + uri.port.to_s + link
		else
			if uri.path =~ /\/$/ 
				return uri.scheme + "://" + uri.host + ":" + uri.port.to_s + uri.path + link
			else
				path_comps = uri.path.split("/")
				path_comps.delete_at(-1)
				return uri.scheme + "://" + uri.host + ":" + uri.port.to_s + ("/" + path_comps.join("/") + "/").gsub(/\/+/, '/') + link
			end
		end
	end
	
	def check_redirect(url, depth)
		aux_uri = URI.parse(url)
		aux_resp = nil
		#begin
			if aux_uri.scheme == "https"
				https_conn = Net::HTTP.new(aux_uri.host, Net::HTTP.https_default_port)
				https_conn.use_ssl = true
				if File.directory?($cert_dir)
					https_conn.ca_path = $cert_dir
					https_conn.verify_mode = OpenSSL::SSL::VERIFY_PEER
					https_conn.verify_depth = 5
				end
				https_conn.ssl_timeout = 10
				# request = Net::HTTP::Get.new(aux_uri.path + "?" + aux_uri.query.to_s)
				# aux_resp = https_conn.request(request)
				$stderr.puts(aux_uri.scheme)
				$stderr.puts(aux_uri.path)
				$stderr.puts(URI.split(url)[7])
				aux_query = URI.split(url)[7]
				https_conn.start { |http|
					# path = @uri.path.to_s
					# path = "/" if @uri.path.to_s == ""
					# request = Net::HTTP::Get.new(aux_uri.path, { 'data' => aux_query.to_s} )
					# aux_resp = Net::HTTP::Get.new(aux_uri.path)
					aux_resp = http.get(aux_uri.path + "?" + aux_query.to_s)
					# @pagecontent = resp.body
				}
				$stderr.puts aux_resp.class.to_s
			else
				aux_resp = Net::HTTP.get_response(aux_uri)
				# $stderr.puts aux_resp.class.to_s
			end
		#rescue Timeout::Error
		#	return nil
		#rescue
		#	return nil
		#end
		# raise FuckingError
		if aux_resp.class.to_s == 'Net::HTTPRedirection' || 
			aux_resp.class.to_s == 'Net::HTTPFound' || 
			aux_resp.class.to_s == 'Net::HTTPMovedPermanently' || 
			aux_resp.class.to_s == 'Net::HTTPTemporaryRedirect'
			# $stderr.puts(url + " Forwards to: " + aux_resp['location'])
			$stderr.puts("BLZ: " + @blz.to_s + " PAGE: " + @url + " (DEPTH " + @depth.to_s + ") REDIR: " + url + " -> " + aux_resp['location'] )
			#another_redir = nil
			#if depth < 1
			#	another_redir = check_redirect(aux_resp['location'], depth + 1)
			#end
			#unless another_redir.nil?
			#	return another_redir
			#else
			# raise FuckingError
			return aux_resp['location']
			#end
		end
		
		return nil
	end
	
	def retry_after_fail
		resp = nil
		if @guessed_url == true
			@url = @url.gsub('https://', 'http://')
			@uri = URI.parse(@url)
			begin
				resp = Net::HTTP.get_response(@uri)
				@pagecontent = resp.body
				# puts resp.class
				# puts resp.body
			rescue
				@timeout = true 
				# $stderr.puts(@url + " ERROR!")
				$stderr.puts("BLZ: " + @blz.to_s + " PAGE: " + @url + " (DEPTH " + @depth.to_s + ") ERRRRRRRRRORRRRRR" )
			end
		else
			@timeout = true 
			$stderr.puts("BLZ: " + @blz.to_s + " PAGE: " + @url + " (DEPTH " + @depth.to_s + ") ERRRRRRRRRORRRRRR" )
		end
		return resp
	end
end